# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"TabNet": "00_core.ipynb",
         "TabNetModel": "00_core.ipynb",
         "SparsemaxFunction": "01_sparsemax.ipynb",
         "Sparsemax": "01_sparsemax.ipynb",
         "Entmax15Function": "01_sparsemax.ipynb",
         "Entmoid15": "01_sparsemax.ipynb",
         "Entmax15": "01_sparsemax.ipynb",
         "sparsemax": "01_sparsemax.ipynb",
         "entmax15": "01_sparsemax.ipynb",
         "entmoid15": "01_sparsemax.ipynb",
         "initialize_non_glu": "02_tab_network.ipynb",
         "initialize_glu": "02_tab_network.ipynb",
         "GBN": "02_tab_network.ipynb",
         "AttentiveTransformer": "02_tab_network.ipynb",
         "FeatTransformer": "02_tab_network.ipynb",
         "GLU_Block": "02_tab_network.ipynb",
         "GLU_Layer": "02_tab_network.ipynb"}

modules = ["core.py",
           "sparsemax.py",
           "tab_network.py"]

doc_url = "https://mgrankin.github.io/fast_tabnet/"

git_url = "https://github.com/mgrankin/fast_tabnet/tree/master/"

def custom_doc_links(name): return None
